/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.data;

import javax.vecmath.AxisAngle4f;
import javax.vecmath.Quat4f;

/**
 * Kierunek skrętu gracza - jeden z 27 możliwych (-1 / 0 / +1) w trzech osiach.
 * Aktualnie kod zakłada interpretację kierunku jako obrót kolejno wokół osi
 * Y, X, Z.
 * Kierunek posiada indeks - liczbę opisującą go, która może służyć np. do
 * przesłania go przez sieć.
 *
 * @author humpolec
 */
public final class Direction {
    // Skręty wokół trzech osi: wartości -1 / 0 / +1.
    // Same kąty obrotu nie są tu przechowywane.
    public final int keyX, keyY, keyZ;

    public int getIndex() {
        return indexOfKeys(keyX, keyY, keyZ);
    }

    /**
     * Policz kwaternion opisujący kierunek dla podanych kątów obrotu wokół
     * osi X, Y, Z.
     */
    public Quat4f getQuat(float angleX, float angleY, float angleZ) {
        Quat4f q = new Quat4f(0, 0, 0, 1), 
            qr = new Quat4f();

        // AxisAngle chce radianów, nie stopni...
        angleX *= Math.PI / 180.0;
        angleY *= Math.PI / 180.0;
        angleZ *= Math.PI / 180.0;
        
        if (keyX != 0 || keyY != 0 || keyZ != 0) { // optymalizacja
            qr.set(new AxisAngle4f(0, 1, 0, keyY * angleX));
            q.mul(qr);
            qr.set(new AxisAngle4f(1, 0, 0, keyX * angleY));
            q.mul(qr);
            qr.set(new AxisAngle4f(0, 0, 1, keyZ * angleZ));
            q.mul(qr);
        }

        return q;
    }

    /**
     * Zwróć kierunek na podstawie skrętów (-1 / 0 / +1) wokół trzech osi.
     */
    public static Direction fromKeys(int keyX, int keyY, int keyZ) {
        return table[indexOfKeys(keyX, keyY, keyZ)];
    }

    public static Direction fromIndex(int i) {
        return table[i];
    }

    public Direction opposite() {
        return fromKeys(-keyX, -keyY, -keyZ);
    }
    
    private Direction(int keyX, int keyY, int keyZ) {
        //assert(fromKeys(keyX, keyY, keyZ) == null);
        this.keyX = keyX;
        this.keyY = keyY;
        this.keyZ = keyZ;
    }
    
    public String toString() {
        return String.format("dir(%2d,%2d,%2d)", keyX, keyY, keyZ);
    }
    
    public int hashCode() {
        return getIndex();
    }

    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        Direction other = (Direction) obj;
        if (getIndex() != other.getIndex())
            return false;
        return true;
    }

    private static int indexOfKeys(int keyX, int keyY, int keyZ) {
        assert (-1 <= keyX && keyX <= 1);
        assert (-1 <= keyY && keyY <= 1);
        assert (-1 <= keyZ && keyZ <= 1);
        return 13 + 9 * keyX + 3 * keyY + keyZ;
    }

    // Stablicowane kierunki skrętów
    private static final Direction[] table = new Direction[27];

    //XXX to jest trochę brzydkie... (porównywanie po referencjach ma działać!)
    public static final Direction STRAIGHT = new Direction(0, 0, 0);

    static {
        for (int x = -1; x <= 1; ++x)
            for (int y = -1; y <= 1; ++y)
                for (int z = -1; z <= 1; ++z)
                    if (!(x==0 && y==0 && z==0))
                        table[indexOfKeys(x, y, z)] = new Direction(x, y, z);
        table[indexOfKeys(0,0,0)] = STRAIGHT;
    }
    
}
