/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.data;

import java.io.IOException;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Field;

import javax.imageio.stream.FileImageOutputStream;

import org.codehaus.jackson.annotate.JsonIgnore;

import netacka3d.data.arena.Mesh;
import netacka3d.data.arena.MeshData;
import netacka3d.data.arena.ObjImporter;
import netacka3d.data.serialization.Serialization;
import netacka3d.misc.FileIO;

/**
 * Ustawienia meczu (mogą być nowe dla każdej rundy).
 * Pola nie są final, ale struktura jest tylko do odczytu 
 * (chyba że wiesz co robisz).
 * 
 * Pola oznaczone @S są do edycji przez użytkownika, po ich modyfikacji 
 * należy zawołać update().
 * 
 * Do testów można konstruować np. tak:
 * <code>
 *   MatchSettings mS = new MatchSettings() {{ 
 *                          collide = false; 
 *                          angleX = 42.0f; 
 *                      }};
 * </code>
 * 
 * @author humpolec
 */
public class MatchSettings implements Cloneable {
    /* @S(...) przed polem ma zawierać info o nazwie, legalnych wartościach itp.
     * Patrz definicja S na końcu klasy.
     */
    // TODO zrobić @S na wszystko co potrzebne
    
    // Czy liczyć kolizje z segmentami (spowalnia)
    public boolean collide = true;
     
    // Nazwa areny (obecnie nazwa pliku .mesh)
    @S(name="arena")
    public String arenaName = "octa";

    // Skala, o jaką powiększamy plik z areną
    @S(name="arena size", min=20.0f, max=500.0f)
    public float arenaSize = 100.0f;
    
    @S(name="speed", min=1.0f, max=15.0f)
    public float speed = 5.0f;
    
    @S(name="turn angle", min=1.0f, max=10.0f)
    public float angles = 5.0f;
    
    // Ilość kroków na sekundę
    @S(name="steps per second", min=0.5f, max=50.0f)
    public float stepRate = 7.0f;
   
    // Wymiary segmentu
    public float segmentLength;
    public float segmentWidth = 5.0f;   
    
    // Kąty skrętu wokół trzech osi
    public float angleX;
    public float angleY;
    public float angleZ;
    
    // Czasy poszczególnych faz meczu (w sekundach)
    public float matchWarmupTime = 1f;
    
    @S(name="round warmup time", min=0.5f, max=5.0f)
    public float roundWarmupTime = 1f;
    
    @S(name="round finish time", min=0.5f, max=5.0f)
    public float roundFinishTime = 1f;
    
    public float matchFinishTime = 1f;    
    
    // Limit punktów na mecz
    @S(name="score limit", min=5, max=50)
    public int scoreLimit = 42;
    
    /** 
     * Uaktualnia dane na podstawie podanych przez użytkownika.
     */
    public void update() {
        if (stepRate >= 5) {
            segmentLength = speed * 4 / stepRate;
            angleX = angleY = angleZ = speed * angles * 6 / stepRate;
        } else {
            segmentLength = 3;
            angleX = angleY = angleZ = 20;
        }
    }
    
    public MatchSettings() {
        update();
    }
    
    /** 
     * Pole z geometrią areny (potrzebne logice i grafice).
     * _Kontrolery_ powinny gwarantować jego załadowanie, za pomocą metod 
     * loadArena() i updateArenaFrom()
     */
    @JsonIgnore
    public MeshData arena; 
    
    public static final String[] arenaNames = new String[] {
        "cube",
        "icosi",
        "octa",
        "cage",
        "torus"
    };
    
    /**
     * Załaduj geometrię areny z pliku do pola arena.
     */
    public void loadArena() throws IOException {
        String fileName = arenaName + ".mesh_bin"; 
        arena = MeshData.loadFrom(FileIO.openFile(fileName));
    }
    
    /**
     * Próbuj skopiować arenę z poprzednich MatchSettings - jak nie ma,
     * załaduj z pliku.
     * 
     * @param mS poprzednie MatchSettings, lub null
     */
    public void updateArenaFrom(MatchSettings mS) throws IOException {
        if (mS != null && mS.arenaName.equals(arenaName))
            arena = mS.arena;
        else
            loadArena();
    }
    
    /**
     * Konstruuje obiekt MatchSettings na podstawie opisu JSON.
     */
    public static MatchSettings fromJson(String s) {
        try {
            return Serialization.mapper.readValue(s, MatchSettings.class);
        } catch (IOException e) {
            return null;
        }
    }
    
    
    /** 
     * Buduje opis obiektu z zaznaczeniem, które pola są niedobre.
     * Prosty przykład użycia refleksji i adnotacji.
     */
    public String pretty() {
        StringBuilder b = new StringBuilder();
        
        try {
            for (Field f: MatchSettings.getFields()) {
                S an = f.getAnnotation(S.class);
                if (an == null)
                    continue;                
                b.append(String.format("%20s: %6s%s\n", 
                        an.name(),
                        f.get(this).toString(),
                        isFieldValid(f) ? "" : " INVALID"));
            }
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
        return b.toString();
    }
    
    public static Field[] getFields() {
        // TODO zapewnić kolejność pól (niby daje w normalnej, ale nie jest to 
        // gwarantowane)
        return MatchSettings.class.getFields();
    }
        
    /** Mówi, czy obiekt ma legalny stan. */
    public boolean isValid() {
        for (Field f: getFields())
            if (!isFieldValid(f))
                return false;
        return true;
    }
        
    /** Mówi, czy dane pole w obiekcie ma legalną wartość. */
    public boolean isFieldValid(Field f) {
        S an = f.getAnnotation(S.class);
        if (an == null)
            return true;
        boolean valid = true;
        try {
            if (f.getType() == float.class) {
                float val = f.getFloat(this);
                valid = (an.min() <= val && val <= an.max());
            }
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
        return valid;
    }
    
    @Retention(RetentionPolicy.RUNTIME)
    public @interface S {
        String name();
        float min() default 0.0f;
        float max() default 1.0f;
    }
    
    public MatchSettings getCopy() {
        try {
            return (MatchSettings)clone();
        } catch (CloneNotSupportedException e) {
            throw new RuntimeException(e);
        }
    }
    
    public static void main(String[] args) {
        System.out.println(new MatchSettings().pretty());
    }
}
