/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.logic;

import java.util.ArrayList;
import java.util.TreeSet;

import netacka3d.misc.Logger;

public class LagCompensation {
    
    private float stepRate;
    
    private long roundStartTime;
    private int lastReceived;
    private ArrayList<Long> segmentTimes;
    private TreeSet<Integer> missingSegments;
    private long whenToResend;
    private long whenAskedForResend = 0;
    private static final int RESEND_TIME = 2000;
    
    public LagCompensation() {
        this(0);
    }
    public LagCompensation(float stepRate) {
        whenToResend = System.currentTimeMillis() + 1000*60*60*24; // jutro
        this.stepRate = stepRate;
        missingSegments = new TreeSet<Integer>();
        segmentTimes = new ArrayList<Long>();
        lastReceived = 0;
    }

    /** 
     * Zwraca czas, który nam pozostał do następnego resenda.
     * Ta metoda nie korzysta ze stepRate.
     */
    public int getTimeout() {
        long now = System.currentTimeMillis();
        return (int) (whenToResend > now ? (whenToResend-now) : 0);
    }
    
    /** 
     * zwraca id pierwszego potrzebnego segmentu, lub -1, jeśli wszystko ok.
     * Ta metoda nie korzysta ze stepRate.
     */
    public int onStep() {
        long now = System.currentTimeMillis();
        if (now >= whenToResend && !missingSegments.isEmpty()) {
            int firstMissing = missingSegments.first();
            whenAskedForResend = now;
            whenToResend = whenAskedForResend + RESEND_TIME;
            assert firstMissing >= 0;
            return firstMissing;
        }
        return -1;
    }

    public void onReceivedSegment(int t) {
        assert this.stepRate > 0;
        if (t > lastReceived) {
            assert !missingSegments.contains(t);
            while (lastReceived < t-1) { // jeśli jest dziura...
                lastReceived++;
                missingSegments.add(lastReceived);
                Logger.d("brakuje mi segmentu "+lastReceived+" (bo dostałem "+t+")");
            }
            lastReceived++;
            assert lastReceived == t;
        }
        else { //dostaliśmy jakiś stary segment
            missingSegments.remove(t);
            Logger.d("dostałem segment "+t+" (stary)");
            if (!missingSegments.isEmpty())
                Logger.d("to teraz brakuje mi od "+missingSegments.first());
        }

        long now = System.currentTimeMillis();
        long expectedOffset = (long) (t * stepRate);
        // jeśli dostaliśmy segment wcześniej, niż się spodziewaliśmy...
        if (roundStartTime + expectedOffset > now)
            roundStartTime = now - expectedOffset; //...regulujemy zegarek
        
        // ustawianie whenToResend
        if (missingSegments.isEmpty()) {
            whenToResend = now + 1000*60*60*24; // jutro
            // (nie za daleko w przyszłości, bo net.work chce inta jako
            // timeout, a nie chcemy, żeby ten int się przepełnił)
        }
        else {
            int firstMissing = missingSegments.first();
            long expectedTime = roundStartTime + 
                   (long) (firstMissing * stepRate);
            whenToResend = RESEND_TIME + Math.max(expectedTime,
                                                  whenAskedForResend);
        }
        
        while (segmentTimes.size() < t+1) 
            segmentTimes.add(null);
        segmentTimes.set(t, now);
        
    }
}