/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.logic;

import java.io.IOException;
import java.util.Map;
import java.util.TreeMap;

import netacka3d.data.Direction;
import netacka3d.data.MatchSettings;
import netacka3d.data.Player;
import netacka3d.data.PlayerColor;
import netacka3d.data.Segment;
import netacka3d.logic.Game.MatchState;
import netacka3d.misc.Logger;
import netacka3d.network.ServerNetworkListener;
import netacka3d.network.ServerNetworkNode;

public class ServerController extends Controller {

    // tego mozna uzywac jako ServerGame, w przeciwienstwie do (publicznego) game
    private ServerGame game; 
    
    private NetListener netListener;
    private ServerNetworkNode net;
    
    private Map<Integer, Direction> dirs;
    
    private int port = 4567;
    
    private int timeUnit; // ile milisekund trwa krok
    private GameThread gameThread;

    private String name;

    private boolean broken = false;
    
    /**
     * Odpala watek gry z podanymi ustawieniami. 
     * Tę metodę wołamy tylko raz. 
     * Działa asynchronicznie, nie blokuje.
     * 
     * @param matchSettings ustawienia nowej gry
     */
    public void startGame(String name, MatchSettings matchSettings, 
                          int port) {
        Logger.i("Rozpoczynanie gry '"+name+"'...");
        
        assert !broken;
        
        this.name = name;
        this.port = port;
        
        assert game == null;
        assert matchSettings != null;
        game = new ServerGame(matchSettings); // to trochę trwa, bo ładuje arenę
        
        dirs = new TreeMap<Integer, Direction>();
        timeUnit = (int) (1000 / matchSettings.stepRate);
        
        gameThread = new GameThread();
        gameThread.setName("S_" + this.name);
        gameThread.start();
    }
    
    /** każe wątkowi gry się zakończyć (działa asynchronicznie) */
    public void endGame() {
        Logger.i("Koniec gry '"+name+"'. Czekanie na zakończenie wątku serwera...");
        assert !broken;
        //assert gameThread.isAlive();
        gameThread.endGame();
    }
    
    /* ---------------------------------------------------------------- */
    
    private static enum State { RUNNING, FINISHED, NETWORK_FAILURE }
    
    protected class GameThread extends Thread {
        
        public volatile ServerController.State state = 
            ServerController.State.RUNNING;
        
        public String errMsg = "";
        
        @Override
        public void run() {
            try {
                go();
            } catch (AssertionError e) {
                broken = true;
                throw e;
            } catch (Exception e) {
                broken = true;
                throw new RuntimeException(e);
            }
        }
            
        private void go() {
            // stawianie sieci
            try { 
                netListener = new NetListener();
                net = new ServerNetworkNode(netListener);
                net.start(port);
            } catch (IOException e) {
                Logger.logException(e);
                notifyStop(e.toString());
                return;
            }

            notifyStart();
            
            // główna pętla!
            long nextStep = System.currentTimeMillis();
            while (state == ServerController.State.RUNNING) {
            	step();
                
                // czekanie (jeśli trzeba)
            	nextStep += timeUnit;
                long milisLeft = nextStep - System.currentTimeMillis();
                if (milisLeft > 0) {
                    try { 
                        Thread.sleep(milisLeft); 
                    }
                    catch (InterruptedException e) {
                        Logger.i("InterruptedException, nie przejmujemy się");
                    }
                }
                else if (milisLeft < -10) {
                    Logger.w("spóźniam się o "+(-milisLeft)+"ms");
                }
            }
        
            if (state == ServerController.State.FINISHED) {
                notifyStop("finished");
                net.shutdown();
            }
            else {
                assert state == ServerController.State.NETWORK_FAILURE;
                notifyStop("network failure: "+errMsg);
            }
        }

        /** 
         * wykonuje pojedynczy krok w Game i powiadamia o zmianach 
         * wszystkich zainteresowanych. 
         */
        private void step() {
            net.work(game);
            
            // krok
            MatchState oldState = game.matchState;
            boolean change = (game.matchState == MatchState.ROUND) ?
                             game.doStep(dirs) : 
                             game.doStep();
            MatchState newState = game.matchState;

            // powiadanianie o zmianie stanu
            if (change)
                net.newGameState(game);
            notifyStep(change, game);
            
            // powiadamianie o segmentach
            //XXX ogólnie to jest syf, bo tak naprawdę to ArenaData powinno
            // pamiętać, które segmenty zostały zmienione
            if ((oldState != newState && newState == MatchState.ROUND_WARMUP) ||
                    (newState == MatchState.ROUND && game.stepNum>0) ||
                    (oldState != newState && newState == MatchState.ROUND_FINISHED))
                broadcastNewSegments();
        }
        
        /** Powiadamia sieć i listenerów o nowych segmentach */
        private void broadcastNewSegments() {
            
            final int t = game.stepNum;
            net.newSegments(t, game.arenaData.getSegments(t));
            
            for (Player p : game.players.getActivePlayers()) {
                Segment seg = game.arenaData.getSegment(p.id, t);
                if (seg != null) {
                    notifySegment(p.id, t, seg);
                }
                else 
                    assert !p.isPlaying() || !p.isAlive(t+1);
            }
        }
        
        /** kończy działanie wątku (do wołania z zewnątrz) */
        public synchronized void endGame() {
            state = ServerController.State.FINISHED;
            interrupt();
        }
    }
    
    /* ---------------------------------------------------------------- */
    
    /** W tej klasie obslugujemy komunikaty przychodzace z sieci.
     * To wszystko bedzie wolane z watku gry, z net.work(). */
    private class NetListener implements ServerNetworkListener {

        private int lastPlayerId = 1;
        
        /**
         * Dodajemy nowego gracza.
         * Pozostali dowiedzą się o nim przy następnej zmianie stanu.
         */
    	@Override
        public int onJoin(String nick, PlayerColor color) {
    	    if (nick.equals("root")) // do testow
    	        return 0;
    	    Player p = new Player(lastPlayerId+1, nick, color);
    	    p = game.players.resolveConflicts(p);
    	    if (p == null) 
    	        return 0;
    	    Logger.i("Dołączyłem gracza "+nick+" z kolorem "+color);
    	    lastPlayerId++;
            game.addPlayer(p);
            dirs.put(lastPlayerId, Direction.STRAIGHT);
            return lastPlayerId;
        }

        @Override
        public void onKey(int id, Direction dir) {
            Logger.d("przyszedl kierunek z sieci: id="+id+", dir="+dir);
            dirs.put(id, dir);
        }

        @Override
        public void onNetFailure(String msg) {
            gameThread.errMsg = msg;
            gameThread.state = ServerController.State.NETWORK_FAILURE;
        }

        @Override
        public void onPart(int id) {
            Logger.i(game.players.getPlayer(id)+" się rozłączył (PART)");
            game.disconnectPlayer(id);
        }
        
    }
}
