/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.network;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;

import netacka3d.data.PlayerColor;
import netacka3d.data.Direction;
import netacka3d.data.MatchSettings;
import netacka3d.data.PlayersTable;
import netacka3d.data.Segment;
import netacka3d.logic.Game;
import netacka3d.logic.Game.MatchState;
import netacka3d.misc.Logger;

/**
 * Moduł sieciowy klienta.
 * @author Leszek
 *
 */
public class ClientNetworkNode {
    /*
     * W ramach last-minute-hacku ClientNetworkNode został przestawiony na
     * synchroniczny java.net. Powód - użycie DatagramChannel.receive() na 
     * Androidzie w sieci wifi nie podłączonej do Internetu powodowało 
     * próbę rezolucji adresu i 20-sekundowy przestój.
     */

    //    private static final String MSG_DISCONNECTED = 
    //        "Utrata połączenia z serwerem";
    private ClientNetworkListener listener;
    //private DatagramChannel channelUDP;
    private DatagramSocket socketUDP;
    private InetSocketAddress serverAddr;

    //private Selector selector;
    private boolean initialized = false; // czy wykonano pomyślnie start
    //    private boolean connected = false; // czy jesteśmy połączeni
    private boolean joined = false; // czy jestesmy w grze

    // @SuppressWarnings("unused")
    private long lastPacketTime; // czas otrzymania ostatniego pakietu

    private final int BUF_SIZE = 6144;
    private ByteBuffer buf = ByteBuffer.allocate(BUF_SIZE);

    private int myId = 0; // id gracza-klienta

    private static final String STRING_UKNOWN_TCPMSG_TYPE = 
        "Unknown TCPMessage type\n";
    private static final String STRING_UKNOWN_UDPMSG_TYPE = 
        "Unknown UDP message type\n";
    //    private static final String STRING_TCPMSG_PARTIALLY_SEND =
    //        "TCPMessage not sent completely\n";
    private static final String STRING_UDP_READ_ERR = 
        "Error while reading from UDP: ";
    private static final String STRING_UDP_WRITE_ERR = 
        "Error while writing to UDP: ";

    // numer ostatniego otrzymanego game state
    private short lastStateNum = -1;

    public ClientNetworkNode(ClientNetworkListener listener) throws IOException 
    {
        this.listener = listener;
        socketUDP = new DatagramSocket(null);
    }



    /**
     * Uruchamia klietna. 
     * @param addr adres serwera
     * @param portUDP port UDP u serwera
     * @return czy udało się dołączyć
     */
    public boolean start(String addr, int portUDP) {
        assert(!initialized);
        serverAddr = new InetSocketAddress(addr,portUDP);
        initialized = true;
        updatePacketTime();
        return true;
    }

    /**
     * Zwalnia przydzielone zasoby m.in. porty
     * @throws IOException 
     */
    private void cleanup() throws IOException {
        if (joined)
            part();
        initialized = false;
        //        connected = false;
        joined = false;
        socketUDP.close();
    }

    /**
     * zakończenie działania sieci w skutek błędu/utraty połączenia
     * @param errmsg
     */
    private void failure(String errmsg) {
        try {
            cleanup();
        } catch (IOException e) {
            errmsg = errmsg + "\nerror during cleanup:" +
            e.getLocalizedMessage();
            Logger.logException(e);
        }
        listener.onNetFailure(errmsg);
    }

    /**
     * kończy działanie serwera
     */
    public void shutdown() {
        assert(initialized);
        try {
            cleanup();
        } catch (IOException e) {
            Logger.logException(e);
        }
    }

    /**
     * sprawdza, czy nie przyszło nic z sieci
     * @param game dotychczasowa wiedza o grze, przyda się samo matchSettings
     * @param timeout ile czasu czekać na komunikat
     */
    public void work(Game game, int timeout) {
        assert initialized;
        assert timeout >= 0;
        
        /* timeout 0 znaczy dla sieci co innego */
        if (timeout == 0)
            timeout = 1;
      
        if (!checkTimeout())
            return;
        try {
            for (;;) {
                buf.clear();
                DatagramPacket p = new DatagramPacket(buf.array(), 0, buf.capacity());
                try {
                    socketUDP.setSoTimeout(timeout);
                    timeout = 1; // tylko za pierwszym razem czekamy długo
                    socketUDP.receive(p);
                } catch (SocketTimeoutException e) {
                    break;
                }
                updatePacketTime();                
                buf.position(0);
                buf.limit(p.getLength());
                byte t = buf.get();
                switch (t) {
                case UDPProtocol.SEGMENTS:
                    if (game.matchSettings == null)
                        break;
                    int stepNum = buf.getInt();
                    Logger.d("KLIENT: Otrzymalem segmenty z chwili " + stepNum);
                    Segment[] segs = 
                        UDPProtocol.getJustSegments(buf,game.matchSettings);
                    assert(buf.remaining() == 0);
                    listener.onNewSegments(stepNum, segs);
                    break;
                case UDPProtocol.NEW_GAME_STATE:
                    short msgNum = buf.getShort();
                    int time = buf.getInt();
                    Logger.i("ODEBRALEM NEWGAMESTATE o num: " 
                            + msgNum + " OD SERWERA DO KLIENTA czas " + time);                        
                    if (msgNum > lastStateNum) {
                        lastStateNum = msgNum;
                        MatchState state = UDPProtocol.getJson(buf,
                                MatchState.class);
                        MatchSettings settings = UDPProtocol.getJson(buf,
                                MatchSettings.class);
                        PlayersTable players = UDPProtocol.getJson(buf,
                                PlayersTable.class);
                        listener.onNewGameState(time, settings, players, state);
                    }
                    if (myId != 0) {
                        Logger.d("WYSYLAM POTWIERDZENIE");
                        buf.clear();
                        buf.put(UDPProtocol.CONFIRM_STATE);
                        buf.putInt(msgNum);
                        buf.flip();
                        sendUDP(buf);
                    }
                    break;
                case UDPProtocol.NOT_JOINED:
                    listener.onJoinFailed();
                    break;
                case UDPProtocol.JOINED:
                    joined = true;
                    myId = buf.getInt();
                    assert myId != 0;
                    //   key(Direction.fromIndex(0));
                    listener.onJoined(myId);
                    break;
                case UDPProtocol.BYE:
                    joined = false;
                    String reason = UDPProtocol.getJson(buf, String.class);
                    Logger.i("OTRZYMALEM BYE " + reason);
                    failure("Connection ended by server:"+reason);
                    return;
                default:
                    Logger.e(STRING_UKNOWN_UDPMSG_TYPE);
                }
            }
        } catch (IOException e) {
            failure(STRING_UDP_READ_ERR
                    + e.getLocalizedMessage());
            //            Logger.logException(e);

        }
        //        Logger.d("work (klient) - koniec");
    }

    private boolean checkTimeout() {
        if (lastPacketTime + UDPProtocol.CLIENT_TIMEOUT <
                System.currentTimeMillis()) {
            Logger.i("CLIENT TIMEOUT");
            if (joined)
                part();
            failure("Client timeout: no packets from server");
            return false;
        }
        return true;
    }

    //   @SuppressWarnings("unused")
    private void updatePacketTime() {
        lastPacketTime = System.currentTimeMillis();
    }



    /**
     * Próba dołączenia się do gry. Jeśli kolor okaże się już zajęty to
     * dostaniemy inny. Jeśli zajęty okaże się nick to też.
     * @param nick nasz nick
     * @param color kolor jaki byśmy chcieli mieć.
     */
    public void join(String nick, PlayerColor color) {
        assert initialized;
        buf.clear();
        buf.put(UDPProtocol.JOIN);
        UDPProtocol.putJson(buf, nick);
        UDPProtocol.putJson(buf, color);
        buf.flip();
        sendUDP(buf);
    }

    private void part() {
        assert joined;
        buf.clear();
        buf.put(UDPProtocol.PART);
        buf.flip();
        joined = false;
        sendUDP(buf);
    }

    /**
     * Wysłanie klawisza do serwera
     * @param dir "klawisz", czyli jaki kierunek chcemy by nasza wstęga obrała
     */
    public void key(Direction dir) {
        assert(joined);
        //        assert(connected);
        assert(dir != null);
        assert(myId != 0);
        Logger.d("KLIENT: wysle sobie klawisz " + dir.getIndex());
        buf.clear();
        buf.put(UDPProtocol.KEY);
        buf.putInt(myId);
        buf.put((byte) dir.getIndex());
        buf.flip();
        sendUDP(buf);
    }

    /**
     * Prośba o nadesłanie jeszcze raz segmentów,
     * które przyszły od chwili t włącznie.
     * @param t time
     */
    public void resend(int t) {
        assert(joined);
        buf.clear();
        buf.put(UDPProtocol.RESEND);
        buf.putInt(t);
        buf.flip();
        sendUDP(buf);
    }

    /**
     * wysyla po UDP bajty z buf.remaining()
     * @param buf bufor z danymi
     * @return true jesli sie udalo, false wpp
     */
    private boolean sendUDP(ByteBuffer buf) {
        try {
            //channelUDP.write(buf);
            DatagramPacket p = new DatagramPacket(buf.array(), buf.position(),
                    buf.remaining(), serverAddr);
            socketUDP.send(p);
        } catch (IOException e) {
            Logger.logException(e);
            failure(STRING_UDP_WRITE_ERR + e.getLocalizedMessage());
            return false;
        }
        return true;
    }




    /**
     * @deprecated
     * Nieuzywane od przejscia na UDPonly
     */
    @SuppressWarnings("unused")
    private void processTCPMessage(TCPMessage msg) throws IOException {
        Logger.i("ODEBRALEM TCPMessage o num: " 
                + msg.num+ " OD SERWERA DO KLIENTA");
        switch (msg.type) {
        case NEW_GAME_STATE:
            listener.onNewGameState(msg.stepNum,
                    msg.settings, msg.players, msg.state);
            if (myId != 0) {
                Logger.i("WYSYLAM POTWIERDZENIE");
                buf.clear();
                buf.put(UDPProtocol.CONFIRM_STATE);
                buf.putInt(msg.num);
                buf.flip();
                sendUDP(buf);
            }
            break;
        case NOT_JOINED:
            listener.onJoinFailed();
            break;
        case JOINED:
            joined = true;
            myId = msg.id;
            assert myId != 0;
            key(Direction.fromIndex(0));
            listener.onJoined(myId);
            break;
        case BYE:
            joined = false;
            Logger.i("OTRZYMALEM BYE " + msg.text);
            failure("Connection ended by server:"+msg.text);
            break;
        default:
            Logger.e(STRING_UKNOWN_TCPMSG_TYPE);
        }
    }

    /**
     * @deprecated
     * Uruchamia klietna. 
     * Na UDPonly bedzie tylko ten drugi start
     * @param addr adres serwera
     * @param portTCP port TCP u serwera
     * @param portUDP port UDP u serwera
     * @param timeout ile maksymalnie milisekund czekać na połączenie
     * @return czy udało się dołączyć
     */
    public boolean start(String addr, int portTCP, int portUDP, int timeout){
        return start(addr,portUDP);
    }
}
