/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.network;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.channels.*;
import java.nio.ByteBuffer;
import java.util.*;

import netacka3d.data.PlayerColor;
import netacka3d.data.Direction;
import netacka3d.data.Segment;
import netacka3d.logic.Game;
import netacka3d.logic.ServerGame;
import netacka3d.logic.Game.MatchState;
import netacka3d.misc.Logger;
import netacka3d.network.TCPMessage.Type;

/**
 * Moduł sieciowy serwera.
 * @author Leszek
 *
 */
public final class ServerNetworkNode {
    private ServerNetworkListener listener;
    private DatagramChannel channelUDP;

    // słownik z adresami UDP graczy
    private Map<Integer,PlayerInfo> players;
    private ArrayList<PlayerInfo> toDisconnect;
    private LinkedList<PlayerInfo> confirmNeeded;

    private boolean initialized = false; // czy wykonano pomyślnie start

    private final int BUF_SIZE = 6144;
    private ByteBuffer buf = ByteBuffer.allocate(BUF_SIZE);
    private ByteBuffer resendBuf = ByteBuffer.allocate(BUF_SIZE);

    private static final String STRING_UKNOWN_TCPMSG_TYPE = 
        "Unknown TCPMessage type\n";
    private static final String STRING_UKNOWN_UDPMSG_TYPE = 
        "Unknown UDP message type\n";
    //    private static final String STRING_TCPMSG_PARTIALLY_SEND =
    //        "TCPMessage not sent completely\n";
    private static final String STRING_UDP_READ_ERR = 
        "Error while reading from UDP: ";
    private static final String STRING_UDP_WRITE_ERR = 
        "Error while writing to UDP: ";

    private short msgNum; // numer wysłanej ostatnio wiadomości (NEW_GAME_STATE)

    public ServerNetworkNode(ServerNetworkListener listener) throws IOException 
    {
        this.listener = listener;
        channelUDP = DatagramChannel.open();
        channelUDP.configureBlocking(false);
        players = new TreeMap<Integer, PlayerInfo>();
        toDisconnect = new ArrayList<PlayerInfo>();
        confirmNeeded = new LinkedList<PlayerInfo>();
    }



    /**
     * Uruchamia serwer. 
     * @param port port na który klienci mają wysyłać klawisze
     */
    public void start(int port) {
        assert(!initialized);
        try {
            channelUDP.socket().bind(new InetSocketAddress(port));  
            initialized = true;
        } catch (IOException e) {
            String errmsg = "Error in server.start:"+e.getLocalizedMessage();
            failure(errmsg);
        }       
    }

    /**
     * Zwalnia przydzielone zasoby m.in. porty
     * @throws IOException 
     */
    private void cleanup() throws IOException {
        //        TCPMessage byeMsg = new TCPMessage();
        //        byeMsg.type = TCPMessage.Type.BYE;
        //        byeMsg.text = "Server closed";
        //        for (PlayerInfo info: players.values()) {
        //            Logger.i("WYSYLAM BYE");
        //            sendTCPMessage(byeMsg, info);
        //        }
        for (PlayerInfo info: players.values()) {
            int id = info.id;
            info.id = 0; // troche hack
            kick(id,"Server shutting down.");
        }
        initialized = false;
        channelUDP.close();
        players.clear();
    }

    /**
     * zakończenie działania sieci w skutek błędu
     * @param errmsg
     */
    private void failure(String errmsg) {
        try {
            cleanup();
        } catch (IOException e) {
            errmsg = errmsg + "\nerror during cleanup:" +
            e.getLocalizedMessage();
            Logger.logException(e);
        }
        listener.onNetFailure(errmsg);
    }

    /**
     * kończy działanie serwera
     */
    public void shutdown() {
        Logger.e("SERVER SHUTDOOOWNNN");
        assert(initialized);

        try {
            cleanup();
        } catch (IOException e) {
            Logger.logException(e);
        }
    }

    private PlayerInfo getPlayerInfo(SocketAddress addr) {
        assert addr != null;
        for (PlayerInfo info: players.values()) {
            if (info.addr.equals(addr))
                return info;
        }
        return new PlayerInfo(addr);
    }

    /**
     * wykonuje kolejno szereg zadań związanych z odbiorem danych z sieci.
     * @param game dane gry
     */
    public void work(ServerGame game) {
        Iterator<PlayerInfo> it = confirmNeeded.iterator();
        while (it.hasNext()) {
            PlayerInfo info = it.next();
            if (validateConfirmation(info)) {
                Logger.d("Nie czekam juz na potwierdzenie od "+info.id);
                it.remove();
            }
        }
        for (PlayerInfo info: toDisconnect) {
            info.disconnect();
        }
        toDisconnect.clear();
        try {
            for (;;) {
                buf.clear();
                SocketAddress addr = channelUDP.receive(buf);
                if (addr == null)
                    break;
                //  PlayerInfo info = getPlayerInfo(addr);
                buf.flip();
                //                Logger.i("COS PRZYSZLOOO");
                try {
                processUDPMessage(buf, addr, game);
                } catch (IOException e) {
                    Logger.e("blad podczas czytania komunikatu UDP");
                    Logger.logException(e);
                }
            }
        } catch (IOException e) {
            failure(STRING_UDP_READ_ERR + e.getLocalizedMessage());
            Logger.logException(e);
        }
    } // work

    /**
     * Sprawdza stan klienta, który miał potwierdzić wiadomość
     * @param info
     * @return true, jeśli klient nie musi już być sprawdzany
     */
    private boolean validateConfirmation(PlayerInfo info) {
        long time = new Date().getTime();
        if (info.id == 0 || info.hasConfirmed) {
            info.resendTimesRemained = UDPProtocol.SERVER_TRIES;
            return true;
        }
        if (time > info.confirmDeadline) {
            if (info.resendTimesRemained == 0) {
                kick(info.id,"SERVER TIMEOUT");
                return true;
            }
            else {
                Logger.i("Wysylam jeszcze raz NEW_GAME_STATE nr " + msgNum +
                        " bo ostatni byl "+info.numToConfirm);
                info.resendTimesRemained--;
                info.confirmDeadline = time + UDPProtocol.SERVER_TIMEOUT;
                resendBuf.rewind();
                sendUDP(resendBuf, info);
            }
        }
        return false;
    }

    /**
     * Przetwarza wiadomość z datagramu UDP
     * @param buf bufor z treścią wiadomości
     * @param addr adres z którego datagram nadszedł
     * @param game dane z grą
     * @throws IOException
     */
    private void processUDPMessage(ByteBuffer buf, SocketAddress addr,
            Game game) throws IOException {
        byte type = buf.get();
        PlayerInfo info = getPlayerInfo(addr);
        switch (type) {

        case UDPProtocol.KEY: // 5 bajtów do przeczytania
            //            Logger.e("to jest KEY! Od: " + addr);
            int id = buf.getInt();
            byte dir = buf.get();
            if (!players.containsKey(id)) {
                Logger.w("nie mam gracza o id " + id +
                        ", wiec olewam ten KEY z addr:" + addr);
                sendBye(info, "sending KEY before being joined");
                break;
            }
            SocketAddress playerAddr = players.get(id).addr;
            if (!addr.equals(playerAddr)) {
                Logger.i("Ktos tu czituje! Wysyla z "+ addr + 
                        " zamiast z " + playerAddr);
                //                Logger.w("adres UDP sie nie zgadza, olewam KEY z addr"+addr);
            }
            listener.onKey(id, Direction.fromIndex(dir));
            break;

        case UDPProtocol.CONFIRM_STATE:
            int recNum = buf.getInt();
            if (!sendByeIfNotJoined(info))
                break;
            Logger.i("Dostalem konf na " + recNum + " od " + addr + "o id "+info.id);
            if (recNum == msgNum)
                info.hasConfirmed = true;
            break;
        case UDPProtocol.JOIN:
            String nick = UDPProtocol.getJson(buf, String.class);
            PlayerColor color = UDPProtocol.getJson(buf, PlayerColor.class);
            int givenId;
            if (info.id == 0)
                givenId = listener.onJoin(nick, color);
            else // już dołączony, po prostu wysyłamy mu komunikat na nowo
                givenId = info.id;
            if (givenId==0) {
                buf.clear();
                buf.put(UDPProtocol.NOT_JOINED);
                buf.flip();
                sendUDP(buf, info);
                Logger.i("wysłałem NOT_JOINED");
            }
            else {
                buf.clear();
                buf.put(UDPProtocol.JOINED);
                buf.putInt(givenId);
                buf.flip();
                Logger.i("przyłączam " + info.addr +
                        " i daje mu id: "+givenId);
                info.joinMe(givenId);
                sendUDP(buf, info);
            }
            break;
        case UDPProtocol.PART:
            info.disconnect();
            break;
        case UDPProtocol.RESEND:
            Logger.i("Ktos chce jeszcze raz segmenty");
            if (game.matchState != MatchState.ROUND) {
                Logger.i("ale mu nie damy, bo jest stan "+game.matchState);
                break;
            }
            if (!sendByeIfNotJoined(info))
                break;
            for (int t = buf.getInt(); t<=game.stepNum; ++t) {
                buf.clear();
                UDPProtocol.newSegments(buf, t, game.arenaData.getSegments(t));
                buf.flip();
                channelUDP.send(buf, addr);
            }
            break;
        case UDPProtocol.GET_GAME_INFO:
            buf.clear();
            buf.put(UDPProtocol.GAME_INFO);
            buf.putInt(game.players.getActivePlayers().size());
            UDPProtocol.putJson(buf, game.matchSettings);
            buf.flip();
            channelUDP.send(buf, addr);
            break;
        default:
            Logger.e(STRING_UKNOWN_UDPMSG_TYPE);
        }        
    }



    private void sendUDP(ByteBuffer buf, PlayerInfo info) {
      //  Logger.i(String.format("Pojdzie %d ", buf.remaining()));
        try {
            channelUDP.send(buf, info.addr);
        } catch (IOException e) {
            toDisconnect.add(info);
            Logger.i("IOException, pewnie ktoś się rozłączył: "
                    + e.getLocalizedMessage());
            return;
        }
    }

    /**
     * wysyła podłączonym klientom (może lepiej graczom, czyli klientom z id!=0)
     * informacje o zmianie stanu gry
     * @param game
     */
    public void newGameState(ServerGame game) {
        //        TCPMessage msg = new TCPMessage();
        //        msg.type = Type.NEW_GAME_STATE;
        //        msg.stepNum = game.stepNum;
        //        msg.players = game.players;
        //        msg.settings = game.matchSettings;
        //        msg.state = game.matchState;
        //        msg.num = TCPMessage.n++;
        resendBuf.clear();
        resendBuf.put(UDPProtocol.NEW_GAME_STATE);
        resendBuf.putShort(++msgNum);
        resendBuf.putInt(game.stepNum);
        UDPProtocol.putJson(resendBuf, game.matchState);
        UDPProtocol.putJson(resendBuf, game.matchSettings);
        UDPProtocol.putJson(resendBuf, game.players);
        resendBuf.limit(resendBuf.position());
        confirmNeeded.clear();
        confirmNeeded.addAll(players.values());
        long confirmDeadline = 
            new Date().getTime() + UDPProtocol.SERVER_TIMEOUT;
        Logger.i("Wysylam NEW_GAME_STATE o num = " + msgNum +
                " do " + players.size() + " graczy");
        for (PlayerInfo info: players.values()) {
            Logger.d("Wysylam newGameState (" + msgNum+") do " +
                    info.addr);
            resendBuf.rewind();
            sendUDP(resendBuf, info);
            info.confirmDeadline = confirmDeadline;
            info.numToConfirm = msgNum;
            info.hasConfirmed = false;
            //            info.msg = msg;
        }
    }

    public void newSegments(int t, Segment[] segs) {
        buf.clear();
        UDPProtocol.newSegments(buf, t, segs);
        buf.limit(buf.position());
        try {
            for (PlayerInfo info:players.values()) {
                buf.rewind();
                //                    Logger.d("Wysylam segment chwili " +
                //                            t + " do " + info.addrUDP);
                channelUDP.send(buf, info.addr);                
            } 
        } catch (IOException e) {
            Logger.logException(e);
            failure(STRING_UDP_WRITE_ERR);
        } 
    }

    /**
     * Wyrzuca gracza o danym id
     * @param id
     */
    public void kick(int id, String reason) {
        assert id != 0;
        assert reason != null;
       // Logger.i("BEDE KICKOWAL");
        buf.clear();
        buf.put(UDPProtocol.BYE);
        UDPProtocol.putJson(buf, reason);
        PlayerInfo info = players.get(id);
        buf.flip();
        sendUDP(buf,info);
        info.disconnect();
    }
    
    private void sendBye(PlayerInfo info, String reason) {
        buf.clear();
        buf.put(UDPProtocol.BYE);
        UDPProtocol.putJson(buf, reason);
        buf.flip();
        sendUDP(buf,info);        
    }
    
    /**
     * 
     * @param info klient do zweryfikowania
     * @return true jesli klient jest graczem,
     *  false jesli nie (wyslano mu BYE).
     */
    private boolean sendByeIfNotJoined(PlayerInfo info) {
        if (info.id == 0) {
            sendBye(info,"sending message before being joined");
            return false;
        }
        return true;
    }


    /**
     * 
     * Klasa przechowująca dane związane z klientem/graczem, potrzebne
     * do komunikacji sieciowej.
     * @author Leszek
     *
     */
    private class PlayerInfo {
        public int id;
        public SocketAddress addr;
        //        public TCPMessage msg; // jaki komunikat musi potwierdzić
        public long confirmDeadline; // do kiedy ma potwierdzić
        public int resendTimesRemained = UDPProtocol.SERVER_TRIES;
        public int numToConfirm = -1; // co ma potwerdzic
        public boolean hasConfirmed = false; // czy potwierdzil

        public PlayerInfo(SocketAddress addr) {
            this.addr = addr;
            this.id = 0;
        }

        void joinMe(int id) {
            this.id = id;
            players.put(id, this);
        }


        /**
         * Rozłącza klienta.
         */
        public void disconnect() {
            Logger.i(String.format("Odłączanie gracza o id %s i adresie %s",
                    id, addr));
            if (id != 0) {
                players.remove(id);
                listener.onPart(id);
                id = 0;
            }
        }
    }

    /**
     * @deprecated od UDPonly nie uzywa sie juz tej metody
     */
    @SuppressWarnings("unused")
    private void processTCPMessage(TCPMessage msg, PlayerInfo info, Game game) {
        Logger.d("ODEBRALEM TCPMessage o id: " + msg.num+ " OD KLIENTA DO SERWERA");
        switch (msg.type) {
        case JOIN:
            int id = listener.onJoin(msg.text, msg.color);
            if (id==0) {
                TCPMessage response = new TCPMessage();
                response.type = Type.NOT_JOINED;
                response.text="BLABLABLA i tak tego nikt nie przetwarza";
                sendTCPMessage(response, info);
                Logger.i("wysłałem NOT_JOINED");
            }
            else {
                TCPMessage response = new TCPMessage();
                response.type = Type.JOINED;
                response.id = id;
                response.players = game.players;
                response.settings = game.matchSettings;
                response.state = game.matchState;
                info.joinMe(id);
                sendTCPMessage(response, info);
                Logger.i("przyłączam " + info.addr +
                        " i daje mu id: "+id);
            }
            break;
        case GET_GAME_INFO:
            TCPMessage response = new TCPMessage();
            response.type = Type.GAME_INFO;
            response.settings = game.matchSettings;
            response.activePlayersCount = 
                game.players.getActivePlayers().size();
            sendTCPMessage(response, info);
            Logger.i("wysłałem GAME_INFO");
            break;
        default:
            Logger.e(STRING_UKNOWN_TCPMSG_TYPE);
        }
    }

    /**
     * @deprecated od UDPonly nie uzywa sie juz tej metody
     */
    private void sendTCPMessage(TCPMessage msg, PlayerInfo info) {
        Logger.d("WYSYLAM TCPMessage o id: " + msg.num+ " DO KLIENTA");
        buf.clear();
        buf.put(UDPProtocol.JSON);
        buf.put(msg.asBytes());
        buf.flip();
        try {
            channelUDP.send(buf, info.addr);
        } catch (IOException e) {
            toDisconnect.add(info);
            Logger.i("IOException, pewnie ktoś się rozłączył: "
                    + e.getLocalizedMessage());
            //Logger.logException(e);
            return;
        }
    }
    
    /**
     * @deprecated
     * Uruchamia serwer. 
     * @param portTCP port na którym przyjmowane są połączenia
     * @param portUDP port na który klienci mają wysyłać klawisze
     * @throws IOException
     */
    public void start(int portTCP, int portUDP){
        start(portUDP);
    }
}