/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.graphics.pc; //#.pc
//#.android package netacka3d.graphics.android;

import javax.vecmath.*;
//import java.nio.ByteBuffer;
//import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.util.ArrayList;

//import java.awt.Container;              //#.pc
import javax.media.opengl.*;            //#.pc
//import javax.media.opengl.glu.GLU;      //#.pc

//import com.sun.opengl.util.Animator;    //#.pc
//import com.sun.opengl.util.FPSAnimator; //#.pc

//#.android import javax.microedition.khronos.egl.EGLConfig;
//#.android import javax.microedition.khronos.opengles.GL10;

//#.android import android.opengl.GLU;
//#.android import android.opengl.GLSurfaceView.Renderer;
//#.android import android.util.Log;

import netacka3d.data.*;
import netacka3d.misc.Logger;

//#replace GL GL10

// XXX przepisać to na wiele buforów
public class RibbonHandler {
    private final int playerId;
    private final PlayerColor color;
    
    private final MatchSettings matchSettings;
    
    private ArrayList<Part> parts;
    // końcówka ostatniego segmentu
    private FloatBuffer lvBuf, lnBuf;
    
    private int size;
    
    private RoundTime death;

    // kolory
    private float[] amb, diff;  
    
    private static final int PART_SIZE = 64;
    private static final int SEGMENT_DATA_SIZE = 4*3;
    
    // bufor pomocniczy
    private float[] tmp;
    
    /* Część wstęgi, wielkości PART_SIZE segmentów. */
    private class Part {
        public FloatBuffer vBuf, nBuf;
        public Segment[] segments;
        
        public Part() {
            vBuf = Util3d.newNativeFloatBuffer(PART_SIZE * SEGMENT_DATA_SIZE);
            nBuf = Util3d.newNativeFloatBuffer(PART_SIZE * SEGMENT_DATA_SIZE);
            segments = new Segment[PART_SIZE];
        }
        
        public void setSegment(int i, Segment seg) {
            assert 0 <= i && i < PART_SIZE;

            if (segments[i] != null && segments[i].equals(seg))
                return;
            
            segments[i] = seg;
            
            Vector3f[] vs = seg.vertices;
            Vector3f[] ns = seg.normals;
    
            vBuf.position(i*SEGMENT_DATA_SIZE);
            nBuf.position(i*SEGMENT_DATA_SIZE);            
            putSegmentData(vBuf, nBuf, vs, ns);
        }
    }
    
    /** Wrzuć dane segmentu do buforów. */
    private void putSegmentData(FloatBuffer vBuf, FloatBuffer nBuf, 
            Vector3f[] vs, Vector3f[] ns) {
        // Używamy tablicy pomocniczej, bo dostęp do direct buffers jest drogi
        for (int k = 0; k < 4; ++k) {
            tmp[k*3+0] = vs[k].x;
            tmp[k*3+1] = vs[k].y;
            tmp[k*3+2] = vs[k].z;
        }
        vBuf.put(tmp);
        
        for (int k = 0; k < 4; ++k) {
            tmp[k*3+0] = ns[k/2].x;
            tmp[k*3+1] = ns[k/2].y;
            tmp[k*3+2] = ns[k/2].z;
        }
        nBuf.put(tmp);
    }
    
    public RibbonHandler(int playerId, PlayerColor color, 
            MatchSettings matchSettings) {
        if (color == null)
            color = PlayerColor.PINK;
        
        this.playerId = playerId;
        this.color = color;
        this.matchSettings = matchSettings;
        
        this.size = 0;
        parts = new ArrayList<Part>();
        parts.add(new Part());
        
        lvBuf = Util3d.newNativeFloatBuffer(SEGMENT_DATA_SIZE);
        lnBuf = Util3d.newNativeFloatBuffer(SEGMENT_DATA_SIZE);
        
        tmp = new float[SEGMENT_DATA_SIZE];
        
        amb = new float[] { color.r*0.5f, color.g*0.5f, color.b*0.5f, 1.0f };
        diff = new float[] { color.r, color.g, color.b, 1.0f };        
    }
    
    public void setSegment(int t, Segment seg) {
        if (t >= size)
            size = t+1;
        
        while (t/PART_SIZE >= parts.size())
            parts.add(new Part());

        parts.get(t/PART_SIZE).setSegment(t%PART_SIZE, seg);
    }
    
    /** 
     * Zwróć segment wstęgi w czasie t - bez ekstrapolacji.
     */
    private Segment getSegmentAt(int t) {
        if (t < size)
            return parts.get(t/PART_SIZE).segments[t%PART_SIZE];
        else
            return null;
    }
    
    /** 
     * Zwróć "czas wstęgi" na podstawie podanego czasu. null oznacza koniec
     * znanej wstęgi.
     */
    private RoundTime getTime(RoundTime rt) {
        if (rt == null) {
            if (death == null)
                return new RoundTime(size-1, 1);
            else
                return death;
        } else if (death != null && rt.isAfter(death))
            return death;
        else
            return rt;
    }
    
    /** 
     * Oblicz nowe położenie i ustaw odpowiednio modelview. Zwraca prawdę, 
     * jeśli się udało. 
     */
    public boolean calcCameraPosition(GL gl, RoundTime rt) {
        rt = getTime(rt);
        
        Segment lastSeg = null;
        
        int t0 = rt.t;
        if (t0 > size-1)
            t0 = size-1;
        for (; t0 >= 0; --t0) {
            lastSeg = getSegmentAt(t0);
            if (lastSeg != null)
                break;
        }
        if (lastSeg == null)
            return false;

        for (; t0 < rt.t; ++t0) {
            lastSeg = lastSeg.calcNext(lastSeg.direction, false, matchSettings);
        }
        
        Vector3f pos = new Vector3f(lastSeg.pos);
        Quat4f heading = new Quat4f(lastSeg.heading);
        
        Vector3f delta = new Vector3f(0, 0, 
                matchSettings.segmentLength*rt.tFrac);
        Util3d.applyQuatToVector(delta, heading);
        pos.add(delta);
        
        heading.interpolate(lastSeg.nextHeading, rt.tFrac);

        heading.inverse();
        Matrix4f mat = new Matrix4f(heading, new Vector3f(), 1.0f);
        gl.glRotatef(180,0,1,0);
        gl.glTranslatef(0.0f,
                -matchSettings.segmentWidth,
                matchSettings.segmentWidth*3.0f);
        gl.glMultMatrixf(Util3d.matrix4fToFloatArray(mat), 0);
        gl.glTranslatef(-pos.x, -pos.y, -pos.z);
        
        return true;
    }
    
    public void endRibbon(RoundTime death) {
        if (size > death.t+1)
            size = death.t+1;
        this.death = death;
    }
    
    /**
     * Wyświetl wstęgę.
     * @param rt czas w jakim jesteśmy, lub null jeśli mamy wyświetlić wszystko
     */
    public void display(GL gl, RoundTime rt) {

        if (size == 0)
            return;
        
        rt = getTime(rt);
        
        int n = rt.t;
        if (n > size)
            n = size;
        Segment lastSeg = getSegmentAt(rt.t);
        if (lastSeg != null) {
            Vector3f[] vs = lastSeg.calcPartVertices(rt.tFrac);
            Vector3f[] ns = lastSeg.calcPartNormals(rt.tFrac);
            
            lvBuf.position(0);
            lnBuf.position(0);
            putSegmentData(lvBuf, lnBuf, vs, ns);                       
        }
                
        gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT, amb, 0);
        gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_DIFFUSE, diff, 0);
                
        gl.glDisable(GL.GL_TEXTURE_2D);
        gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
        
        int pos = 0;
        for (Part p: parts) {
            p.vBuf.position(0);
            p.nBuf.position(0);
            gl.glVertexPointer(3, GL.GL_FLOAT, 0, p.vBuf);
            gl.glNormalPointer(GL.GL_FLOAT, 0, p.nBuf);
            
            int m;
            if (pos + PART_SIZE <= n)
                m = PART_SIZE;
            else
                m = n - pos;
            pos += PART_SIZE;
            
            for (int i = 0; i < m; ++i)
                gl.glDrawArrays(GL.GL_TRIANGLE_FAN, 4*i, 4);
        }
            
        if (lastSeg != null) {
            lvBuf.position(0);
            lnBuf.position(0);
            gl.glVertexPointer(3, GL.GL_FLOAT, 0, lvBuf);
            gl.glNormalPointer(GL.GL_FLOAT, 0, lnBuf);
            gl.glDrawArrays(GL.GL_TRIANGLE_FAN, 0, 4);
        }
        
        gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
    }
}
