/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.graphics.pc; //#.pc
//#.android package netacka3d.graphics.android;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;

import javax.media.opengl.GL; //#.pc

//#.android import javax.microedition.khronos.egl.EGLConfig;
//#.android import javax.microedition.khronos.opengles.GL10;
//#.android import android.opengl.GLU;

import netacka3d.data.Util3d;
import netacka3d.misc.FileIO;

//#replace GL GL10
//#replace glTexParameteri glTexParameterx
//#replace glOrtho glOrthof

public class TextRenderer {
    private ByteBuffer tBuf;
    private FloatBuffer vBuf;
    private FloatBuffer tcBuf;

    private Texture texture;
    
    private int screenW, screenH;
    private float rows, cols;
    
    public static final int TEXTURE_WIDTH = 256;
    public static final int TEXTURE_HEIGHT = 512;
    public static final float FONT_WIDTH = 14.5f;
    public static final float FONT_HEIGHT = 28;
    public static final int TEXT_ROWS = 20;
    public static final int TEXT_COLS = 40;
    
    public void init(GL gl) {
        texture = new Texture(gl, "font.png");
        assert (texture.w == TEXTURE_WIDTH);
        assert (texture.h == TEXTURE_HEIGHT);
        
        vBuf = Util3d.newNativeFloatBuffer(4*3);
        vBuf.put(new float[] { 0, 0, 0, 
                               0, 1, 0, 
                               1, 1, 0,
                               1, 0, 0 });
        tcBuf = Util3d.newNativeFloatBuffer(4*2);
        tcBuf.put(new float[] { 0, 0,
                                0, 1,
                                1, 1,
                                1, 0 });
    }
    
    public void destroy(GL gl) {
        texture.destroy(gl);
    }
    
    public void setScreenDimensions(int w, int h) {
        screenW = w;
        screenH = h;
    }
    
    public void startFrame(GL gl) {
        gl.glDisable(GL.GL_DEPTH_TEST);
        gl.glDisable(GL.GL_LIGHTING);
        
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glPushMatrix();
        gl.glLoadIdentity();
        
        float ratio = (screenW/(float)FONT_WIDTH)/(screenH/(float)FONT_HEIGHT);
        if (ratio < TEXT_COLS/(float)TEXT_ROWS) {
            cols = TEXT_COLS;
            rows = TEXT_COLS/ratio;
        } else {
            rows = TEXT_ROWS;
            cols = TEXT_ROWS*ratio;
        }
        
        gl.glOrtho(0, cols, rows, 0, -1, 1);
        
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();  
        gl.glLoadIdentity();
        
        gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glLoadIdentity();
        gl.glScalef(FONT_WIDTH/(float)TEXTURE_WIDTH, 
                FONT_HEIGHT/(float)TEXTURE_HEIGHT, 1);
        
        gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
        gl.glEnable(GL.GL_TEXTURE_2D);
        gl.glBindTexture(GL.GL_TEXTURE_2D, texture.num);
        tcBuf.position(0);
        vBuf.position(0);
        gl.glTexCoordPointer(2, GL.GL_FLOAT, 0, tcBuf);
        gl.glVertexPointer(3, GL.GL_FLOAT, 0, vBuf);
        
        gl.glEnable(GL.GL_BLEND);
        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
        
        gl.glMatrixMode(GL.GL_MODELVIEW);
    }
    
    public void setColor(GL gl, float r, float g, float b) {
        gl.glColor4f(r, g, b, 1);
    }
    
    public void drawChar(GL gl, int x, int y, char c) {
        int i = (int) c;
        if (i > 255)
            return;

        int px = i%16, py = i/16;

        //gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glTranslatef(x, y, 0);
        
        gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glTranslatef(px, py, 0);
        
        gl.glDrawArrays(GL.GL_TRIANGLE_FAN, 0, 4);
        
        //gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glTranslatef(-px, -py, 0);
                
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glTranslatef(-x, -y, 0);        
    }
    
    public void drawString(GL gl, int x, int y, String s) {
        for (int i = 0; i < s.length(); ++i)
            drawChar(gl, x+i, y, s.charAt(i));
    }
    
    public void drawStringCenter(GL gl, int y, String s) {
        drawString(gl, (int)(cols-s.length())/2, y, s);
    }
    
    public void drawStringRight(GL gl, int x, int y, String s) {
        drawString(gl, x-s.length()+1, y, s);
    }
    
    public void endFrame(GL gl) {
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glEnable(GL.GL_LIGHTING);
        
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glPopMatrix();

        gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glLoadIdentity();
        
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPopMatrix();
        
        gl.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY);
        gl.glDisable(GL.GL_TEXTURE_2D);        
        
        gl.glDisable(GL.GL_BLEND);
    }
}
